/**
  Rendering demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2010-04-08 (Public Domain)
*/
#include <iostream>
#include <cmath>
#include <GL/glew.h> /* 
OpenGL Extensions Wrangler:  for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <GL/glut.h> /* OpenGL Utilities Toolkit, for GUI tools */
#include "ogl/glsl.h"
#include "ogl/glsl.cpp"
#include "ogl/minicam.h"
#include "osl/mat4.h"
#include "osl/mat4_inverse.h"

#include <vector>


class triangle {
public:
	int a[3]; // 1-based index into points and normals array
	triangle(int *a_) {
		for (int i=0;i<3;i++) a[i]=a_[i];
	}
};

void display(void) 
{
	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE); // turn on Z writes
	glEnable(GL_BLEND);
	
	glClearColor(0.4,0.5,0.7,0.0); // background color
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); // flush any ancient matrices
	gluPerspective(70.0, // fov
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT),
		0.1,
		1000.0); // z clipping planes
	float cameraVelocity=3.0;
	oglCameraLookAt(cameraVelocity);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity(); // flush any ancient matrices
	
	/* The proxy sphere should be gently rotating */
	const float rotate_rate=0.01; /* revolutions per second */
	float time_uniform=glutGet(GLUT_ELAPSED_TIME)*0.001;
	
/* Load and draw the cat model */
	GLhandleARB catshader=MAKE_PROGRAM_OBJECT(
		varying vec3 position; // location of fur-deformed geometry
		varying vec3 texcoords; // location of *original* geometry
		varying float fur; // distance along fur: 0.0=skin; 1.0=tips
		varying vec3 normal; // surface normal
		void main(void) {
			vec4 world=gl_ModelViewMatrix*gl_Vertex;
			position=vec3(world);
			normal=vec3(gl_NormalMatrix*gl_Normal);
			gl_Position=gl_ProjectionMatrix*world;
			texcoords=vec3(gl_ModelViewMatrix*gl_MultiTexCoord0);
			fur=gl_MultiTexCoord0.w;
		}
	,
		varying vec3 position; // location of fur-deformed geometry
		varying vec3 texcoords; // location of *original* geometry
		varying float fur; // distance along fur: 0.0=skin; 1.0=tips
		varying vec3 normal; // surface normal
		
		vec4 hair_color(vec3 pos) {
			vec3 follicle=floor(pos+0.5); // center of follicle: perturb too?
			float radius=length(pos-follicle);
			vec4 hair_color=vec4(0.7,0.3,0.2,1.0-pow(fur,3.0)); /* hairs */
			vec4 bg_color=vec4(0.3,0.2,0.1,pow(1.0-fur,2.0)); /* background */
			
			float hair_frac=0.0;
			if (radius<0.4) hair_frac=1.0;
			
			return bg_color+hair_frac*(hair_color-bg_color); // linear blend
		}
		
		void main(void) {
			vec4 color=hair_color(texcoords*100.0);
		
			const vec3 L=normalize(vec3(0.4,0.0,1.0));
			vec3 N=normalize(normal);
			float d=clamp(dot(N,L),0.0,0.8)+0.2;
			gl_FragColor=color*vec4(vec3(d),1.0);
		}
	);
	glUseProgramObjectARB(catshader);
	
	
	std::vector<vec3> points; points.push_back(vec3(-999.0));
	std::vector<vec3> normals; normals.push_back(vec3(-999.0));
	std::vector<vec3> texcoords; texcoords.push_back(vec3(-999.0));
	std::vector<triangle> tris;
	FILE *f=fopen("cat_lopoly.obj","r");
	char line[1000];
	while (NULL!=fgets(line,1000,f)) {
		switch (line[0]) {
		case 'v': {
			float x,y,z;
			sscanf(line,"v%f%f%f",&x,&y,&z);
			vec3 p=0.5*vec3(x,y,z);
			texcoords.push_back(p);
			p.z+=0.02*sin(10.0*p.x + time_uniform); // rippling fur
			points.push_back(p);
			normals.push_back(vec3(0.0,0.0,0.0));
			break;
		}
		case 'f': {
			int a[3];
			sscanf(line,"f%d%d%d",&a[0],&a[1],&a[2]);
			vec3 N=normalize(cross(points[a[0]]-points[a[1]],
				points[a[0]]-points[a[2]]));
			for (int i=0;i<3;i++) normals[a[i]]+=N;
			tris.push_back(triangle(a));
			break;
		}
		default: /* ignore that line */ ;
		}
	}
	fclose(f);
	
	/* normalize the normals */
	for (int n=1;n<(int)normals.size();n++) normals[n]=normalize(normals[n]);
	
	/* Draw many alpha-blended layers of fur */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	float furdist=0.2;
	for (float fur=0.0;fur<=1.0;fur+=1.0/16.0)  //<- looks much better with 32 layers
	{
		if (fur>0.0) glDepthMask(GL_FALSE); // turn off depth writes
		glBegin(GL_TRIANGLES); 
		for (int t=0;t<(int)tris.size();t++)
		{
			for (int i=0;i<3;i++) {
				int vertex=tris[t].a[i];
				// The normal is used for lighting
				glNormal3fv(normals[vertex]);

				// The texcoords determine the hair roots
				vec3 p=texcoords[vertex];
				glTexCoord4fv(vec4(p.x,p.y,p.z,fur));

				// The geometry shows where stuff lands onscreen
				p=points[vertex];
				p+=(fur*furdist)*normals[vertex];
				glVertex3fv(p);
			}
		}
		glEnd();
	}
	glDepthMask(GL_TRUE);
	
	glUseProgramObjectARB(0);
	
	glutPostRedisplay(); // continual animation
	glutSwapBuffers();
	
/* Estimate framerate */
	static double last_time=0.0;
	static int framecount=0, last_framecount=0;
	framecount++;
	double cur_time=0.001*glutGet(GLUT_ELAPSED_TIME);
	if (cur_time>0.5+last_time) 
	{ // every half a second, estimate fps
		double time=cur_time-last_time;
		double time_per_frame=time/(framecount-last_framecount);
		char str[100];
		sprintf(str,"Fuzzy Renderer: %.1f fps, %.1f ms/frame",
			1.0/time_per_frame,1.0e3*time_per_frame);
		glutSetWindowTitle(str);
		last_framecount=framecount;
		last_time=cur_time;
	}
}

int main(int argc,char *argv[]) 
{
	glutInit(&argc,argv);
	glutInitDisplayMode(GLUT_RGBA + GLUT_DEPTH + GLUT_DOUBLE);
	glutInitWindowSize(800,600);
	glutCreateWindow("Demo Application");
	
	glewInit();
	
	glutDisplayFunc(display);
	oglCameraInit();
	camera=vec3(0.0,-2.0,0.0);
	correct_head_tilt=true; // allow free flight
	
	glutMainLoop();
	return 0;
}
